﻿#LoiLoNote School Sideloading Setup Version 1.0
param(
    [ValidateSet('Select', 'Install','Uninstall','ShowLog','ProvisionedInstall','ProvisionedUninstall', 'ElevatedProvisionedInstall', 'ElevatedProvisionedUninstall')]
    [string] $Action = 'Select',
    [string] $Packages,
    [switch] $Always,
    [switch] $Force
)

$ErrorActionPreference = "Stop"

Set-PSDebug -Strict

$DefaultPackageName = "LoiLoinc.51956F723A3DD.s1"
$CompanyName = "LoiLo"
$ScriptPath = (Get-Variable MyInvocation).Value.MyCommand.Path
$ScriptDirPath = Split-Path $ScriptPath -Parent
$ScriptName = $(Get-ChildItem $ScriptPath).BaseName
$WorkDirPath = Join-Path "$env:LOCALAPPDATA" (Join-Path $CompanyName $ScriptName)
$TempDirPath = Join-Path $WorkDirPath "Temp"
$LogDirPath = Join-Path $WorkDirPath "Logs"

#region Log

$LogMaxLines = 1000
$LogFileName = "log.txt"
$LogFilePath = Join-Path $LogDirPath $LogFileName
$LogBackupFileName = "log.txt.back"
$LogBackupFilePath = Join-Path $LogDirPath $LogBackupFileName

function Append-Log{

    param(
        [ValidateSet("INFO", "ERROR", ignorecase=$true)]
        [string] $type,
        [string] $msg
    )

    $now = Get-Date -Format "yyyy/MM/dd HH:mm:ss"
    $message = ("$now {0,-5} $msg" -f $type)

    if ($type -ne "ERROR") {
        Write-Host $message
    } else {
        Write-Host $message -ForegroundColor Red
    }

    if(-not (Test-Path $LogDirPath -PathType Container)){
        New-Item $LogDirPath -ItemType Directory -Force | Out-Null
    }

    if(Test-Path $LogFilePath -PathType Leaf){
        $line = (Get-Content $LogFilePath | Measure-Object –Line).Lines
        if( $line -ge $LogMaxLines ){
            if(Test-Path $LogBackupFilePath -PathType Leaf){
                Remove-Item $LogBackupFilePath -Force 
            }        
            Rename-Item $LogFilePath -NewName $LogBackupFileName 
        }
    }

    Write-Output $message | Out-File ${LogFilePath} -Append -Encoding utf8
}

#endregion

#region Logic

function Find-File {
    param( [string] $dirPath, [string] $filePattern)

    $queryPath = (Join-Path $dirPath "*.*")
    $foundItems = @(Get-ChildItem $queryPath -Include $filePattern | Where-Object { $_.Mode -NotMatch "d" })
    if($foundItems.Length -le 0) {
        return $null
    }
    return $foundItems[0]
}

function Get-TargetPackageInfo {
    param([string] $dirName)
    $packageJsonFilePath = Join-Path $dirName "package.json"
    $packageJsonContent = Get-Content -Path $packageJsonFilePath -Encoding UTF8 -Raw 
    $info = ConvertFrom-Json $packageJsonContent
    return $info
}

function Compare-Version {
    param([string] $version1, [string] $version2)

    $array1 = @($version1 -split "." , 0 , "simplematch")
    $array2 = @($version2 -split "." , 0 , "simplematch")

    for($i = 0; $i -lt 4; $i++) {
        $contains1 = ($array1.Count - 1) -ge $i;
        $contains2 = ($array2.Count - 1) -ge $i;
        if((-not $contains1) -and (-not $contains2)) {
            return 0;
        } elseif(-not $contains1) {
            return -1;
        } elseif(-not $contains2) {
            return 1;
        }

        $num1 = [int]$array1[$i]
        $num2 = [int]$array2[$i]

        if($num1 -gt $num2) {
            return 1;
        } elseif($num1 -lt $num2) {
            return -1;
        }
    }

    return 0;
}

function Get-DependencyPackages {
    param([string] $dirPath)

    $dependenciesDirPath = (Join-Path $dirPath "Dependencies")
    $dependencyPackages = @()
    if (Test-Path $dependenciesDirPath) {
        $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "*.appx") | Where-Object { $_.Mode -NotMatch "d" }

        if (($Env:Processor_Architecture -eq "x86" -or $Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x86"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x86\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        } 
        if (($Env:Processor_Architecture -eq "amd64") -and (Test-Path (Join-Path $dependenciesDirPath "x64"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "x64\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
        if (($Env:Processor_Architecture -eq "arm") -and (Test-Path (Join-Path $dependenciesDirPath "arm"))) {
            $dependencyPackages += Get-ChildItem (Join-Path $dependenciesDirPath "arm\*.appx") | Where-Object { $_.Mode -NotMatch "d" }
        }
    }

    return $dependencyPackages
}

function Install-Package {
    param([string] $dirPath)
    
    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    if ($dependencyPackages.FullName.Count -gt 0) {
        Append-Log INFO "Found dependency($dependencyPackages)"
        Add-AppxPackage "$bundleFilePath" -DependencyPath $dependencyPackages.FullName -ForceApplicationShutdown
    } else {
        Add-AppxPackage "$bundleFilePath" -ForceApplicationShutdown
    }
}

function Install-ProvisionedPackage {
    param([string] $dirPath)
    
    $bundleFilePath = Find-File $dirPath "*.appxbundle"
    if($bundleFilePath -eq $null) {
        throw "No bundle file in package."
    }

    $dependencyPackages = Get-DependencyPackages $dirPath

    if ($dependencyPackages.FullName.Count -gt 0) {
        Append-Log INFO "Found dependency($dependencyPackages)"
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -DependencyPackagePath $dependencyPackages.FullName -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    } else {
        $image = Add-AppxProvisionedPackage -Online -PackagePath "$bundleFilePath" -SkipLicense
        Append-Log INFO ($image | Format-List | Out-String)
    }
}

function Uninstall-Package {
    param([string] $packageName)
    $package = Get-AppxPackage -Name $packageName
    if($package) {
        Append-Log INFO "Uninstalling..."
        Remove-AppxPackage -Package $package.PackageFullName
        Append-Log INFO "Complete!"
    } else {
        Append-Log INFO "Not installed."
    }
}

function Uninstall-ProvisionedPackage{
    param([string] $packageName)
    $packages = @(Get-AppxProvisionedPackage -Online | Where DisplayName -eq $packageName)
    if($packages) {
        foreach( $package in $packages ) {
            $packageFullname = $package.PackageName
            Append-Log INFO "Uninstalling ProvisionedPackage($packageFullname)..."
            $image = Remove-AppxProvisionedPackage -Online -PackageName $packageFullname
            Append-Log INFO ($image | Format-List | Out-String)
        }
        Append-Log INFO "Complete!" 
    } else {
        Append-Log INFO "Not installed."
    }   
}

function Run-Install {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Not exists $indexFilePath ."
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
        throw "Not exists $targetDirPath ."
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    $targetVersion = $targetPackageInfo.Version
    $targetPublisherId = $targetPackageInfo.PublisherId

    $installedPackage = Get-AppxPackage -Name $packageName

    $removeRequired = $false
    if($installedPackage) {
        $compareResult = Compare-Version $installedPackage.Version $targetVersion
        if($compareResult -lt 0) {
            Append-Log INFO ("Currently installed $packageName Version=" + $installedPackage.Version + "(Older).")

            $installedPackagePublisherId = $installedPackage.PublisherId
            if($installedPackagePublisherId -ne $targetPublisherId) {
                Append-Log INFO "PublisherId changed.($installedPackagePublisherId to $targetPublisherId)"
                $removeRequired = $true
            }
        } elseif($compareResult -gt 0) {
            Append-Log INFO ("Currently installed $packageName Version=" + $installedPackage.Version + "(Newer).")
            if(-not $Always) {
                return
            }
            $removeRequired = $true
        } else {
            Append-Log INFO ("Already installed $packageName Version=" + $installedPackage.Version + ".")
            return
        }               
    } else {
        Append-Log INFO "Currently not installed $packageName."
    }

    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    if( $removeRequired ) {
        Append-Log INFO "Removing Installed Version..."
        Uninstall-Package $packageName
    }

    Append-Log INFO "Installing Version=$targetVersion ..."
    Install-Package $copyDirPath
    Append-Log INFO "Complete!"
}

function Run-ProvisionedInstall {
    param([string] $packageDirPath)

    $indexFilePath = (Join-Path $packageDirPath "index.txt")
    if(-not(Test-Path $indexFilePath)) {
        throw "Not exists $indexFilePath ."
    }

    $indexFileContent = (Get-Content $indexFilePath) -as [string[]]
    if($indexFileContent.Length -le 0) {
        throw "No entry in $indexFilePath"
    }

    $targetDirName = $indexFileContent[0]
    $targetDirPath = Join-Path $packageDirPath $targetDirName
    if(-not(Test-Path $targetDirPath)) {
        throw "Not exists $targetDirPath ."
    }

    $targetPackageInfo = Get-TargetPackageInfo $targetDirPath
    $packageName = $targetPackageInfo.Name
    Uninstall-ProvisionedPackage $packageName

    $copyDirPath = Join-Path $TempDirPath $targetDirName
    Copy-Item -Path $targetDirPath -Destination $copyDirPath -Force -Recurse | Out-Null

    Append-Log INFO "Installing ProvisionedPackage..."
    Install-ProvisionedPackage $copyDirPath
    Append-Log INFO "Complete!"
}

function Ensure-TempDir {
    Append-Log INFO "Ready $TempDirPath"
    if(Test-Path $TempDirPath)
    {
        Remove-Item $TempDirPath -Recurse -Force
    }
    New-Item $TempDirPath -ItemType Directory -Force | Out-NUll
}

function Run{

    $running = $Action;
    if($running -eq 'Select') {
        $input = [int](Read-Host "Please select option below.`n0:Exit, 1:Install, 2:Uninstall, 3:ShowLog, 4:ProvisionedInstall, 5:ProvisionedUninstall`n=>")
        if($input -eq 1) {
            $running = 'Install';
        } elseif ($input -eq 2) {
            $running = 'Uninstall';
        } elseif ($input -eq 3) {
            $running = 'ShowLog';
        } elseif ($input -eq 4) {
            $running = 'ProvisionedInstall';
        } elseif ($input -eq 5) {
            $running = 'ProvisionedUninstall';
        } else {
            Append-Log INFO "Exit."
            return $true
        }
    }

    Append-Log INFO "Action=$running"

    if($running -eq 'Install') {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath)"
        Run-Install $packageDirPath

        return $false
    } elseif ($running -eq 'Uninstall') {
        Uninstall-Package $DefaultPackageName

        return $false
    } elseif ($running -eq 'ShowLog') {
        Start-Process "$LogDirPath" | Out-Null

        return $true
    } elseif ($running -eq 'ProvisionedInstall') {

        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedInstall'

        if($Force) {
            $relaunchArgs += ' -Force'
        }

        if( $Packages -ne "" ) {
            $relaunchArgs += ' -Packages "' + $Packages + '"'
        }
        
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq 'ProvisionedUninstall') {
        $relaunchArgs = '-ExecutionPolicy Unrestricted -file "' + $ScriptPath + '" -Action ElevatedProvisionedUninstall'
        Start-Process "$PsHome\PowerShell.exe" -Verb RunAs -ArgumentList $relaunchArgs -PassThru | Out-Null

        return $true
    } elseif($running -eq "ElevatedProvisionedInstall") {
        Ensure-TempDir
        $packageDirPath = Join-Path $ScriptDirPath "Packages"
        if( $Packages -ne "" ){
            $packageDirPath = $Packages
        }
        Append-Log INFO "Param( Always=$Always, Packages=$packageDirPath)"
        Run-ProvisionedInstall $packageDirPath

        return $false
    } elseif ($running -eq "ElevatedProvisionedUninstall"){
        Uninstall-ProvisionedPackage $DefaultPackageName

        return $false
    }

    return $false
}

#endregion


#
# Main script entry point
#

Append-Log INFO "[Start] $ScriptName"

$SilentExit = $false
try {
    $SilentExit = Run
} catch {
    Append-Log ERROR $_.Exception.Message
} finally {
    Append-Log INFO "[End] $ScriptName"
    if ((-not $SilentExit) -and (-not $Force)) {
        Pause
    }
}

# SIG # Begin signature block
# MIIafwYJKoZIhvcNAQcCoIIacDCCGmwCAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUYiOUgMzwviziuMgvNcmglvgA
# CLagghVuMIIEkzCCA3ugAwIBAgIQR4qO+1nh2D8M4ULSoocHvjANBgkqhkiG9w0B
# AQUFADCBlTELMAkGA1UEBhMCVVMxCzAJBgNVBAgTAlVUMRcwFQYDVQQHEw5TYWx0
# IExha2UgQ2l0eTEeMBwGA1UEChMVVGhlIFVTRVJUUlVTVCBOZXR3b3JrMSEwHwYD
# VQQLExhodHRwOi8vd3d3LnVzZXJ0cnVzdC5jb20xHTAbBgNVBAMTFFVUTi1VU0VS
# Rmlyc3QtT2JqZWN0MB4XDTEwMDUxMDAwMDAwMFoXDTE1MDUxMDIzNTk1OVowfjEL
# MAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UE
# BxMHU2FsZm9yZDEaMBgGA1UEChMRQ09NT0RPIENBIExpbWl0ZWQxJDAiBgNVBAMT
# G0NPTU9ETyBUaW1lIFN0YW1waW5nIFNpZ25lcjCCASIwDQYJKoZIhvcNAQEBBQAD
# ggEPADCCAQoCggEBALw1oDZwIoERw7KDudMoxjbNJWupe7Ic9ptRnO819O0Ijl44
# CPh3PApC4PNw3KPXyvVMC8//IpwKfmjWCaIqhHumnbSpwTPi7x8XSMo6zUbmxap3
# veN3mvpHU0AoWUOT8aSB6u+AtU+nCM66brzKdgyXZFmGJLs9gpCoVbGS06CnBayf
# UyUIEEeZzZjeaOW0UHijrwHMWUNY5HZufqzH4p4fT7BHLcgMo0kngHWMuwaRZQ+Q
# m/S60YHIXGrsFOklCb8jFvSVRkBAIbuDlv2GH3rIDRCOovgZB1h/n703AmDypOmd
# RD8wBeSncJlRmugX8VXKsmGJZUanavJYRn6qoAcCAwEAAaOB9DCB8TAfBgNVHSME
# GDAWgBTa7WR0FJwUPKvdmam9WyhNizzJ2DAdBgNVHQ4EFgQULi2wCkRK04fAAgfO
# l31QYiD9D4MwDgYDVR0PAQH/BAQDAgbAMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/
# BAwwCgYIKwYBBQUHAwgwQgYDVR0fBDswOTA3oDWgM4YxaHR0cDovL2NybC51c2Vy
# dHJ1c3QuY29tL1VUTi1VU0VSRmlyc3QtT2JqZWN0LmNybDA1BggrBgEFBQcBAQQp
# MCcwJQYIKwYBBQUHMAGGGWh0dHA6Ly9vY3NwLnVzZXJ0cnVzdC5jb20wDQYJKoZI
# hvcNAQEFBQADggEBAMj7Y/gLdXUsOvHyE6cttqManK0BB9M0jnfgwm6uAl1IT6TS
# IbY2/So1Q3xr34CHCxXwdjIAtM61Z6QvLyAbnFSegz8fXxSVYoIPIkEiH3Cz8/dC
# 3mxRzUv4IaybO4yx5eYoj84qivmqUk2MW3e6TVpY27tqBMxSHp3iKDcOu+cOkcf4
# 2/GBmOvNN7MOq2XTYuw6pXbrE6g1k8kuCgHswOjMPX626+LB7NMUkoJmh1Dc/VCX
# rLNKdnMGxIYROrNfQwRSb+qz0HQ2TMrxG3mEN3BjrXS5qg7zmLCGCOvb4B+MEPI5
# ZJuuTwoskopPGLWR5Y0ak18frvGm8C6X0NL2KzwwggV0MIIEXKADAgECAhAnZu5W
# 60nzjqvXcKL8hN4iMA0GCSqGSIb3DQEBDAUAMG8xCzAJBgNVBAYTAlNFMRQwEgYD
# VQQKEwtBZGRUcnVzdCBBQjEmMCQGA1UECxMdQWRkVHJ1c3QgRXh0ZXJuYWwgVFRQ
# IE5ldHdvcmsxIjAgBgNVBAMTGUFkZFRydXN0IEV4dGVybmFsIENBIFJvb3QwHhcN
# MDAwNTMwMTA0ODM4WhcNMjAwNTMwMTA0ODM4WjCBhTELMAkGA1UEBhMCR0IxGzAZ
# BgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9yZDEaMBgG
# A1UEChMRQ09NT0RPIENBIExpbWl0ZWQxKzApBgNVBAMTIkNPTU9ETyBSU0EgQ2Vy
# dGlmaWNhdGlvbiBBdXRob3JpdHkwggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIK
# AoICAQCR6FSS0gpWsawNJN3Fz0RndJkrN6N9I3AAcbxT38T6KhKPS38QVr2fcHK3
# YX/JSw8Xpz3jsARh7v8Rl8f0hj4K+j5c+ZPmNHrZFGvnnLOFoIJ6dq9xkNfs/Q36
# nGz637CC9BR++b7Epi9Pf5l/tfxnQ3K9DADWietrLNPtj5gcFKt+5eNu/Nio5JIk
# 2kNrYrhV/erBvGy2i/MOjZrkm2xpmfh4SDBF1a3hDTxFYPwyllEnvGfDyi62a+pG
# x8cgoLEfZd5ICLqkTqnyg0Y3hOvozIFIQ2dOciqbXL1MGyiKXCJ7tKuY2e7gUYPD
# CUZObT6Z+pUX2nwzV0E8jVHtC7ZcryxjGt9XyD+86V3Em69FmeKjWiS0uqlWPc9v
# qv9JWL7wqP/0uK3pN/u6uPQLOvnoQ0IeidiEyxPx2bvhiWC4jChWrBQdnArncevP
# Dt09qZahSL0896+1DSJMwBGB7FY79tOi4lu3sgQiUpWAk2nojkxl8ZEDLXB0AuqL
# ZxUpaVICu9ffUGpVRr+goyhhf3DQw6KqLCGqR84onAZFdr+CGCe01a60y1Dma/RM
# hnEw6abfFobg2P9A3fvQQoh/ozM6LlweQRGBY84YcWsr7KaKtzFcOmpH4MN5WdYg
# Gq/yapiqcrxXStJLnbsQ/LBMQeXtHT1eKJ2czL+zUdqnR+WEUwIDAQABo4H0MIHx
# MB8GA1UdIwQYMBaAFK29mHo0tCb3+sQmVO8DveAky1QaMB0GA1UdDgQWBBS7r34C
# Pfqm8TyEjq3uOJjs2TIy1DAOBgNVHQ8BAf8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB
# /zARBgNVHSAECjAIMAYGBFUdIAAwRAYDVR0fBD0wOzA5oDegNYYzaHR0cDovL2Ny
# bC51c2VydHJ1c3QuY29tL0FkZFRydXN0RXh0ZXJuYWxDQVJvb3QuY3JsMDUGCCsG
# AQUFBwEBBCkwJzAlBggrBgEFBQcwAYYZaHR0cDovL29jc3AudXNlcnRydXN0LmNv
# bTANBgkqhkiG9w0BAQwFAAOCAQEAZL+D8V+ahdDNuKEpVw3oWvfR6T7ydgRu8VJw
# ux48/00NdGrMgYIl08OgKl1M9bqLoW3EVAl1x+MnDl2EeTdAE3f1tKwc0DurFxLW
# 7zQYfivpedOrV0UMryj60NvlUJWIu9+FV2l9kthSynOBvxzz5rhuZhEFsx6ULX+R
# lZJZ8UzOo5FxTHxHDDsLGfahsWyGPlyqxC6Cy/kHlrpITZDylMipc6LrBnsjnd6i
# 801Vn3phRZgYaMdeQGsj9Xl674y1a4u3b0b0e/E9SwTYk4BZWuBBJB2yjxVgWEfb
# 725G/RX12V+as9vYuORAs82XOa6Fux2OvNyHm9Gm7/E7bxA4bzCCBXcwggRfoAMC
# AQICEFalkVib5C/irmWR5v1Sd7swDQYJKoZIhvcNAQELBQAwfTELMAkGA1UEBhMC
# R0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4GA1UEBxMHU2FsZm9y
# ZDEaMBgGA1UEChMRQ09NT0RPIENBIExpbWl0ZWQxIzAhBgNVBAMTGkNPTU9ETyBS
# U0EgQ29kZSBTaWduaW5nIENBMB4XDTE1MDEyMzAwMDAwMFoXDTE3MDEyMjIzNTk1
# OVowgcUxCzAJBgNVBAYTAkpQMREwDwYDVQQRDAgyMzEtMDAwNjEeMBwGA1UECAwV
# T3V0c2lkZSBVbml0ZWQgU3RhdGVzMRQwEgYDVQQHDAtZb2tvaGFtYXNoaTEMMAoG
# A1UECQwDNDAzMSMwIQYDVQQJDBpOYWtha3UgTWluYW1pbmFrYWRvcmkgNCA0MzES
# MBAGA1UECgwJTG9pTG8gaW5jMRIwEAYDVQQLDAlMb2lMbyBpbmMxEjAQBgNVBAMM
# CUxvaUxvIGluYzCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAOd9CM8+
# xWoMTb7Khv5y0RKZNWAELx1N5UdIefdqVPFzHSyvsOCnVgEB+nUKebApdFhKTdkq
# GitQwtNcFu6s+49rdVOPG7ZI6emkPSG1+68DUIaKw4I9c3kPV1CcnCrQAzNnTj2e
# wDKwiXKy1dtUPiwHGztO6AT5h2OGGwRfjK8DbWr+6psl73I5oBDarMf7mbvTMhop
# ze1IAaazYqBoenfOpiqgEM4Mx01+lBffGxMois1mvHIn/DlrrdZ6W0xgTcRh1FcV
# KYE7lEKP3DxshbKGhM8gVNJ82A7Y7Fgrp1Z52BIqNu/IL2HnahJRIincWvjgVLzF
# ztov3I+zWd6FYe0CAwEAAaOCAagwggGkMB8GA1UdIwQYMBaAFCmRYP+KTfrr+aZq
# uM/55ku9Sc4SMB0GA1UdDgQWBBRcaJm/RsASokRd/yL3zztIMFdl0zAOBgNVHQ8B
# Af8EBAMCB4AwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAKBggrBgEFBQcDAzARBglg
# hkgBhvhCAQEEBAMCBBAwRgYDVR0gBD8wPTA7BgwrBgEEAbIxAQIBAwIwKzApBggr
# BgEFBQcCARYdaHR0cHM6Ly9zZWN1cmUuY29tb2RvLm5ldC9DUFMwQwYDVR0fBDww
# OjA4oDagNIYyaHR0cDovL2NybC5jb21vZG9jYS5jb20vQ09NT0RPUlNBQ29kZVNp
# Z25pbmdDQS5jcmwwdAYIKwYBBQUHAQEEaDBmMD4GCCsGAQUFBzAChjJodHRwOi8v
# Y3J0LmNvbW9kb2NhLmNvbS9DT01PRE9SU0FDb2RlU2lnbmluZ0NBLmNydDAkBggr
# BgEFBQcwAYYYaHR0cDovL29jc3AuY29tb2RvY2EuY29tMBkGA1UdEQQSMBCBDmxv
# aWxvQGxvaWxvLnR2MA0GCSqGSIb3DQEBCwUAA4IBAQCHzuwnDpHP1aRHjEfsY7s6
# CDrSogzi6c8aa+e7otQeiTet9279Uv+BKwKLkCRBxmhNCyIVJhSt9/Pwlg6b8ubR
# tPxIXaCVDmvy0+h0iB+j9f/tK/DumZM7yTGkRuH+hnl1Wwgce5d0OVLXjRV1pmPo
# WB5mwoaCsQIbxjwBzGpASV7BQ7JEdbbnFHJaPVHfJPP9OG/y6tT6wT/gQsQ8slyb
# NSkR7HOl8jvP5y2Iolu7bXNctxX1HXzBRKuwlXUoBJv8pQ0j3keUh6gQKDuXO+oQ
# yaDwL3G/F37SrfWUV3Xbp0+/vXRmVuKtcAi4u1QqGvBzm5UcHli5++sLCwkI6O3J
# MIIF4DCCA8igAwIBAgIQLnyHzA6TSlL+lP0ct800rzANBgkqhkiG9w0BAQwFADCB
# hTELMAkGA1UEBhMCR0IxGzAZBgNVBAgTEkdyZWF0ZXIgTWFuY2hlc3RlcjEQMA4G
# A1UEBxMHU2FsZm9yZDEaMBgGA1UEChMRQ09NT0RPIENBIExpbWl0ZWQxKzApBgNV
# BAMTIkNPTU9ETyBSU0EgQ2VydGlmaWNhdGlvbiBBdXRob3JpdHkwHhcNMTMwNTA5
# MDAwMDAwWhcNMjgwNTA4MjM1OTU5WjB9MQswCQYDVQQGEwJHQjEbMBkGA1UECBMS
# R3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdTYWxmb3JkMRowGAYDVQQKExFD
# T01PRE8gQ0EgTGltaXRlZDEjMCEGA1UEAxMaQ09NT0RPIFJTQSBDb2RlIFNpZ25p
# bmcgQ0EwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCmmJBjd5E0f4rR
# 3elnMRHrzB79MR2zuWJXP5O8W+OfHiQyESdrvFGRp8+eniWzX4GoGA8dHiAwDvth
# e4YJs+P9omidHCydv3Lj5HWg5TUjjsmK7hoMZMfYQqF7tVIDSzqwjiNLS2PgIpQ3
# e9V5kAoUGFEs5v7BEvAcP2FhCoyi3PbDMKrNKBh1SMF5WgjNu4xVjPfUdpA6M0ZQ
# c5hc9IVKaw+A3V7Wvf2pL8Al9fl4141fEMJEVTyQPDFGy3CuB6kK46/BAW+QGiPi
# XzjbxghdR7ODQfAuADcUuRKqeZJSzYcPe9hiKaR+ML0btYxytEjy4+gh+V5MYnmL
# Agaff9ULAgMBAAGjggFRMIIBTTAfBgNVHSMEGDAWgBS7r34CPfqm8TyEjq3uOJjs
# 2TIy1DAdBgNVHQ4EFgQUKZFg/4pN+uv5pmq4z/nmS71JzhIwDgYDVR0PAQH/BAQD
# AgGGMBIGA1UdEwEB/wQIMAYBAf8CAQAwEwYDVR0lBAwwCgYIKwYBBQUHAwMwEQYD
# VR0gBAowCDAGBgRVHSAAMEwGA1UdHwRFMEMwQaA/oD2GO2h0dHA6Ly9jcmwuY29t
# b2RvY2EuY29tL0NPTU9ET1JTQUNlcnRpZmljYXRpb25BdXRob3JpdHkuY3JsMHEG
# CCsGAQUFBwEBBGUwYzA7BggrBgEFBQcwAoYvaHR0cDovL2NydC5jb21vZG9jYS5j
# b20vQ09NT0RPUlNBQWRkVHJ1c3RDQS5jcnQwJAYIKwYBBQUHMAGGGGh0dHA6Ly9v
# Y3NwLmNvbW9kb2NhLmNvbTANBgkqhkiG9w0BAQwFAAOCAgEAAj8COcPu+Mo7id4M
# bU2x8U6ST6/COCwEzMVjEasJY6+rotcCP8xvGcM91hoIlP8l2KmIpysQGuCbsQci
# GlEcOtTh6Qm/5iR0rx57FjFuI+9UUS1SAuJ1CAVM8bdR4VEAxof2bO4QRHZXavHf
# WGshqknUfDdOvf+2dVRAGDZXZxHNTwLk/vPa/HUX2+y392UJI0kfQ1eD6n4gd2HI
# TfK7ZU2o94VFB696aSdlkClAi997OlE5jKgfcHmtbUIgos8MbAOMTM1zB5TnWo46
# BLqioXwfy2M6FafUFRunUkcyqfS/ZEfRqh9TTjIwc8Jvt3iCnVz/RrtrIh2IC/gb
# qjSm/Iz13X9ljIwxVzHQNuxHoc/Li6jvHBhYxQZ3ykubUa9MCEp6j+KjUuKOjswm
# 5LLY5TjCqO3GgZw1a6lYYUoKl7RLQrZVnb6Z53BtWfhtKgx/GWBfDJqIbDCsUgmQ
# Fhv/K53b0CDKieoofjKOGd97SDMe12X4rsn4gxSTdn1k0I7OvjV9/3IxTZ+evR5s
# L6iPDAZQ+4wns3bJ9ObXwzTijIchhmH+v1V04SF3AwpobLvkyanmz1kl63zsRQ55
# ZmjoIs2475iFTZYRPAmK0H+8KCgT+2rKVI2SXM3CZZgGns5IW9S1N5NGQXwH3c/6
# Q++6Z2H/fUnguzB9XIDj5hY5S6cxggR7MIIEdwIBATCBkTB9MQswCQYDVQQGEwJH
# QjEbMBkGA1UECBMSR3JlYXRlciBNYW5jaGVzdGVyMRAwDgYDVQQHEwdTYWxmb3Jk
# MRowGAYDVQQKExFDT01PRE8gQ0EgTGltaXRlZDEjMCEGA1UEAxMaQ09NT0RPIFJT
# QSBDb2RlIFNpZ25pbmcgQ0ECEFalkVib5C/irmWR5v1Sd7swCQYFKw4DAhoFAKB4
# MBgGCisGAQQBgjcCAQwxCjAIoAKAAKECgAAwGQYJKoZIhvcNAQkDMQwGCisGAQQB
# gjcCAQQwHAYKKwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwIwYJKoZIhvcNAQkE
# MRYEFKbn16sEzRxg3B2b07yDrfq2ljKNMA0GCSqGSIb3DQEBAQUABIIBACFKcyzF
# lUrsGPDBpFd0TlGugoYVTRLio1B8/I2p8ESzTpcqR2OggQezw8eGWJoefaHFf2tE
# PODjWqpXZ+3+bpbuzjx/xJhEgUquYCZ6S0IuVwJbFXrfwdWu0kna7jN+7VG5ezzz
# ZKrOoY8zUpzqcOtIcObF8gcjWosmVTzbFgrfhURvjkIszvUKai/dTYhZtGFkWngl
# GGfTqOLcfcS+PpM+5rR4n2WJNka6BZXkR3G3y8bxXmJx+JkfABjMxbdrMbAav9RH
# cec4MX+ZcMJFd4Cj7pUWOwwU0bEsASNwupXmH87Dtt0G2Jklkktf86TTWMY9FKfp
# BB/ScKSOFcP6kO6hggJEMIICQAYJKoZIhvcNAQkGMYICMTCCAi0CAQAwgaowgZUx
# CzAJBgNVBAYTAlVTMQswCQYDVQQIEwJVVDEXMBUGA1UEBxMOU2FsdCBMYWtlIENp
# dHkxHjAcBgNVBAoTFVRoZSBVU0VSVFJVU1QgTmV0d29yazEhMB8GA1UECxMYaHR0
# cDovL3d3dy51c2VydHJ1c3QuY29tMR0wGwYDVQQDExRVVE4tVVNFUkZpcnN0LU9i
# amVjdAIQR4qO+1nh2D8M4ULSoocHvjAJBgUrDgMCGgUAoF0wGAYJKoZIhvcNAQkD
# MQsGCSqGSIb3DQEHATAcBgkqhkiG9w0BCQUxDxcNMTUwMzE4MDY0OTM0WjAjBgkq
# hkiG9w0BCQQxFgQUIsHpH6WvtNG/5xsascHex4Ae41wwDQYJKoZIhvcNAQEBBQAE
# ggEAWdsZt5DKfcRp0lbdCYZ+HWU7XtTQ5ZAPC/wUj/UF7KwSmD4zJJ6I6PKPn9De
# xQODrH5lRiOj4zBaPpEkXoH/JpgAZTOr3h3HkxPulUlJPqZVdnbbLguSZCAYVwVc
# v2y4uRiOFi3K6Pp/Cj7YTDLfgu1gRDIPf7lSsDNnYc+4zfcYJrv7vwJpqhuhsOby
# q9FPjZW0qzNyXgJKFnhu7UoCl2eOSHymJgvlf00rqJcl5yfwsLbwCaw5d7ORn/fz
# i4LvBG3sJ+9Jdc4Xi7fRjvPNTeqvTjGqdW0akjrbaVNtGGVzcU9EnBLv2OTIhhSS
# Ce4ewtBLWXbTOnvfYNIcLbCbOg==
# SIG # End signature block
